@@ -1,4 +1,6 @@
 """Utility functions related to file operations."""
+from __future__ import annotations
+
 import copy
 import logging
 import os
@@ -6,11 +8,11 @@ import pathlib
 import subprocess
 import sys
 from argparse import Namespace
-from collections import OrderedDict
+from collections import OrderedDict, defaultdict
 from contextlib import contextmanager
 from pathlib import Path
 from tempfile import NamedTemporaryFile
-from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Set, Union, cast
+from typing import TYPE_CHECKING, Any, Iterator, cast
 
 # import wcmatch
 import wcmatch.pathlib
@@ -44,13 +46,16 @@ def abspath(path: str, base_dir: str) -> str:
     return os.path.normpath(path)
 
 
-def normpath(path: Union[str, BasePathLike]) -> str:
+def normpath(path: str | BasePathLike) -> str:
     """
     Normalize a path in order to provide a more consistent output.
 
     Currently it generates a relative path but in the future we may want to
     make this user configurable.
     """
+    # prevent possible ValueError with relpath(), when input is an empty string
+    if not path:
+        path = "."
     # conversion to string in order to allow receiving non string objects
     relpath = os.path.relpath(str(path))
     path_absolute = os.path.abspath(str(path))
@@ -61,7 +66,7 @@ def normpath(path: Union[str, BasePathLike]) -> str:
 
 
 @contextmanager
-def cwd(path: Union[str, BasePathLike]) -> Iterator[None]:
+def cwd(path: str | BasePathLike) -> Iterator[None]:
     """Context manager for temporary changing current working directory."""
     old_pwd = os.getcwd()
     os.chdir(path)
@@ -80,7 +85,7 @@ def expand_path_vars(path: str) -> str:
     return path
 
 
-def expand_paths_vars(paths: List[str]) -> List[str]:
+def expand_paths_vars(paths: list[str]) -> list[str]:
     """Expand the environment or ~ variables in a list."""
     paths = [expand_path_vars(p) for p in paths]
     return paths
@@ -132,15 +137,18 @@ class Lintable:
 
     def __init__(
         self,
-        name: Union[str, Path],
-        content: Optional[str] = None,
-        kind: Optional[FileType] = None,
+        name: str | Path,
+        content: str | None = None,
+        kind: FileType | None = None,
     ):
         """Create a Lintable instance."""
         # Filename is effective file on disk, for stdin is a namedtempfile
         self.filename: str = str(name)
         self.dir: str = ""
-        self.kind: Optional[FileType] = None
+        self.kind: FileType | None = None
+        self.stop_processing = False  # Set to stop other rules from running
+        self._data: Any = None
+        self.line_skips: dict[int, set[str]] = defaultdict(set)
 
         if isinstance(name, str):
             self.name = normpath(name)
@@ -183,6 +191,7 @@ class Lintable:
 
         # determine base file kind (yaml, xml, ini, ...)
         self.base_kind = kind_from_path(self.path, base=True)
+        self.abspath = self.path.absolute()
 
     def __getitem__(self, key: Any) -> Any:
         """Provide compatibility subscriptable support."""
@@ -195,12 +204,19 @@ class Lintable:
     def get(self, key: Any, default: Any = None) -> Any:
         """Provide compatibility subscriptable support."""
         try:
-            return self.__getitem__(key)
+            return self[key]
         except NotImplementedError:
             return default
 
     def _populate_content_cache_from_disk(self) -> None:
-        self._content = self.path.resolve().read_text(encoding="utf-8")
+        # Can raise UnicodeDecodeError
+        try:
+            self._content = self.path.resolve().read_text(encoding="utf-8")
+        except FileNotFoundError as ex:
+            if vars(options).get("progressive"):
+                self._content = ""
+            else:
+                raise ex
         if self._original_content is None:
             self._original_content = self._content
 
@@ -261,7 +277,7 @@ class Lintable:
 
     def __hash__(self) -> int:
         """Return a hash value of the lintables."""
-        return hash((self.name, self.kind))
+        return hash((self.name, self.kind, self.abspath))
 
     def __eq__(self, other: object) -> bool:
         """Identify whether the other object represents the same rule match."""
@@ -273,9 +289,29 @@ class Lintable:
         """Return user friendly representation of a lintable."""
         return f"{self.name} ({self.kind})"
 
+    @property
+    def data(self) -> Any:
+        """Return loaded data representation for current file, if possible."""
+        if not self._data:
+            if str(self.base_kind) == "text/yaml":
+                from ansiblelint.utils import (  # pylint: disable=import-outside-toplevel
+                    parse_yaml_linenumbers,
+                )
+
+                self._data = parse_yaml_linenumbers(self)
+                # Lazy import to avoid delays and cyclic-imports
+                if "append_skipped_rules" not in globals():
+                    # pylint: disable=import-outside-toplevel
+                    from ansiblelint.skip_utils import append_skipped_rules
+
+                self._data = append_skipped_rules(self._data, self)
+
+            # will remain None if we do not know how to load that file-type
+        return self._data
+
 
 # pylint: disable=redefined-outer-name
-def discover_lintables(options: Namespace) -> Dict[str, Any]:
+def discover_lintables(options: Namespace) -> dict[str, Any]:
     """Find all files that we know how to lint."""
     # git is preferred as it also considers .gitignore
     git_command_present = [
@@ -291,14 +327,14 @@ def discover_lintables(options: Namespace) -> Dict[str, Any]:
 
     try:
         out_present = subprocess.check_output(
-            git_command_present, stderr=subprocess.STDOUT, universal_newlines=True
+            git_command_present, stderr=subprocess.STDOUT, text=True
         ).split("\x00")[:-1]
         _logger.info(
             "Discovered files to lint using: %s", " ".join(git_command_present)
         )
 
         out_absent = subprocess.check_output(
-            git_command_absent, stderr=subprocess.STDOUT, universal_newlines=True
+            git_command_absent, stderr=subprocess.STDOUT, text=True
         ).split("\x00")[:-1]
         _logger.info("Excluded removed files using: %s", " ".join(git_command_absent))
 
@@ -332,7 +368,7 @@ def strip_dotslash_prefix(fname: str) -> str:
     return fname[2:] if fname.startswith("./") else fname
 
 
-def guess_project_dir(config_file: Optional[str]) -> str:
+def guess_project_dir(config_file: str | None) -> str:
     """Return detected project dir or current working directory."""
     path = None
     if config_file is not None:
@@ -344,9 +380,8 @@ def guess_project_dir(config_file: Optional[str]) -> str:
         try:
             result = subprocess.run(
                 ["git", "rev-parse", "--show-toplevel"],
-                stderr=subprocess.PIPE,
-                stdout=subprocess.PIPE,
-                universal_newlines=True,
+                capture_output=True,
+                text=True,
                 check=True,
             )
 
@@ -373,7 +408,7 @@ def guess_project_dir(config_file: Optional[str]) -> str:
     return path
 
 
-def expand_dirs_in_lintables(lintables: Set[Lintable]) -> None:
+def expand_dirs_in_lintables(lintables: set[Lintable]) -> None:
     """Return all recognized lintables within given directory."""
     should_expand = False
 
