@@ -2,16 +2,22 @@
 from __future__ import annotations
 
 import logging
+import re
 import sys
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
+from collections import namedtuple
+from typing import TYPE_CHECKING, Any
 
+import black
 import jinja2
+from ansible.errors import AnsibleError, AnsibleFilterError, AnsibleParserError
 from ansible.parsing.yaml.objects import AnsibleUnicode
+from jinja2.exceptions import TemplateSyntaxError
 
+from ansiblelint.constants import LINE_NUMBER_KEY
 from ansiblelint.file_utils import Lintable
 from ansiblelint.rules import AnsibleLintRule
 from ansiblelint.skip_utils import get_rule_skips_from_line
-from ansiblelint.utils import LINE_NUMBER_KEY, parse_yaml_from_file
+from ansiblelint.utils import parse_yaml_from_file, template
 from ansiblelint.yaml_utils import nested_items_path
 
 if TYPE_CHECKING:
@@ -19,6 +25,9 @@ if TYPE_CHECKING:
 
 
 _logger = logging.getLogger(__package__)
+KEYWORDS_WITH_IMPLICIT_TEMPLATE = ("changed_when", "failed_when", "until", "when")
+
+Token = namedtuple("Token", "lineno token_type value")
 
 
 class JinjaRule(AnsibleLintRule):
@@ -28,8 +37,11 @@ class JinjaRule(AnsibleLintRule):
     severity = "LOW"
     tags = ["formatting"]
     version_added = "v6.5.0"
+    _ansible_error_re = re.compile(
+        r"^(?P<error>.*): (?P<detail>.*)\. String: (?P<string>.*)$", flags=re.MULTILINE
+    )
 
-    env = jinja2.Environment()
+    env = jinja2.Environment(trim_blocks=False)
     _tag2msg = {
         "invalid": "Syntax error in jinja2 template: {value}",
         "spacing": "Jinja2 spacing could be improved: {value} -> {reformatted}",
@@ -39,12 +51,76 @@ class JinjaRule(AnsibleLintRule):
         """Generate error message."""
         return self._tag2msg[tag].format(value=value, reformatted=reformatted)
 
-    def matchtask(
-        self, task: Dict[str, Any], file: Optional[Lintable] = None
-    ) -> Union[bool, str, MatchError]:
-        for _, v, _ in nested_items_path(task):
+    # pylint: disable=too-many-branches
+    def matchtask(  # noqa: C901
+        self, task: dict[str, Any], file: Lintable | None = None
+    ) -> bool | str | MatchError:
+        for key, v, _ in nested_items_path(task):
             if isinstance(v, str):
-                reformatted, details, tag = self.check_whitespace(v)
+
+                try:
+                    template(
+                        basedir=file.dir if file else ".",
+                        value=v,
+                        variables={},
+                        fail_on_error=True,  # we later decide which ones to ignore or not
+                    )
+                # ValueError RepresenterError
+                except AnsibleError as exc:
+                    bypass = False
+                    orig_exc = exc.orig_exc if getattr(exc, "orig_exc", None) else exc
+                    match = self._ansible_error_re.match(
+                        getattr(orig_exc, "message", str(orig_exc))
+                    )
+                    if (
+                        isinstance(exc, AnsibleFilterError)
+                        or "unable to locate collection" in orig_exc.message
+                    ):
+                        bypass = True
+                    elif re.match(
+                        r"^the template file (.*) could not be found for the lookup$",
+                        orig_exc.message,
+                    ) or re.match(r"could not locate file in lookup", orig_exc.message):
+                        bypass = True
+                    elif isinstance(orig_exc, AnsibleParserError):
+                        # "An unhandled exception occurred while running the lookup plugin '...'. Error was a <class 'ansible.errors.AnsibleParserError'>, original message: Invalid filename: 'None'. Invalid filename: 'None'"
+
+                        # An unhandled exception occurred while running the lookup plugin 'template'. Error was a <class 'ansible.errors.AnsibleError'>, original message: the template file ... could not be found for the lookup. the template file ... could not be found for the lookup
+
+                        # ansible@devel (2.14) new behavior:
+                        # AnsibleError(TemplateSyntaxError): template error while templating string: Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'. String: Foo {{ buildset_registry.host | ipwrap }}. Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'
+                        bypass = True
+                    elif (
+                        isinstance(orig_exc, (AnsibleError, TemplateSyntaxError))
+                        and match
+                    ):
+                        error = match.group("error")
+                        detail = match.group("detail")
+                        # string = match.group("string")
+                        if error.startswith("template error while templating string"):
+                            bypass = False
+                        elif detail.startswith("unable to locate collection"):
+                            _logger.debug("Ignored AnsibleError: %s", exc)
+                            bypass = True
+                        else:
+                            # breakpoint()
+                            bypass = False
+                    elif re.match(r"^lookup plugin (.*) not found$", exc.message):
+                        # lookup plugin 'template' not found
+                        bypass = True
+
+                    # AnsibleFilterError: 'obj must be a list of dicts or a nested dict'
+                    # AnsibleError: template error while templating string: expected token ':', got '}'. String: {{ {{ '1' }} }}
+                    # AnsibleError: template error while templating string: unable to locate collection ansible.netcommon. String: Foo {{ buildset_registry.host | ipwrap }}
+                    if not bypass:
+                        return self.create_matcherror(
+                            message=str(exc),
+                            linenumber=task[LINE_NUMBER_KEY],
+                            filename=file,
+                            tag=f"{self.id}[invalid]",
+                        )
+
+                reformatted, details, tag = self.check_whitespace(v, key=key)
                 if reformatted != v:
                     return self.create_matcherror(
                         message=self._msg(tag=tag, value=v, reformatted=reformatted),
@@ -55,18 +131,18 @@ class JinjaRule(AnsibleLintRule):
                     )
         return False
 
-    def matchyaml(self, file: Lintable) -> List[MatchError]:
+    def matchyaml(self, file: Lintable) -> list[MatchError]:
         """Return matches for variables defined in vars files."""
-        data: Dict[str, Any] = {}
-        raw_results: List[MatchError] = []
-        results: List[MatchError] = []
+        data: dict[str, Any] = {}
+        raw_results: list[MatchError] = []
+        results: list[MatchError] = []
 
         if str(file.kind) == "vars":
             data = parse_yaml_from_file(str(file.path))
             # pylint: disable=unused-variable
-            for k, v, path in nested_items_path(data):
+            for key, v, path in nested_items_path(data):
                 if isinstance(v, AnsibleUnicode):
-                    reformatted, details, tag = self.check_whitespace(v)
+                    reformatted, details, tag = self.check_whitespace(v, key=key)
                     if reformatted != v:
                         results.append(
                             self.create_matcherror(
@@ -90,14 +166,16 @@ class JinjaRule(AnsibleLintRule):
             results.extend(super().matchyaml(file))
         return results
 
-    def lex(self, text: str) -> List[Tuple[int, str, str]]:
+    def lex(self, text: str) -> list[Token]:
         """Parse jinja template."""
         # https://github.com/pallets/jinja/issues/1711
         self.env.keep_trailing_newline = True
 
         self.env.lstrip_blocks = False
         self.env.trim_blocks = False
-        tokens = list(self.env.lex(text))
+        tokens = [
+            Token(lineno=t[0], token_type=t[1], value=t[2]) for t in self.env.lex(text)
+        ]
         new_text = self.unlex(tokens)
         if text != new_text:
             _logger.debug(
@@ -105,7 +183,7 @@ class JinjaRule(AnsibleLintRule):
             )
         return tokens
 
-    def unlex(self, tokens: List[Any]) -> str:
+    def unlex(self, tokens: list[Token]) -> str:
         """Return original text by compiling the lex output."""
         result = ""
         last_lineno = 1
@@ -120,207 +198,100 @@ class JinjaRule(AnsibleLintRule):
 
     # pylint: disable=too-many-branches,too-many-statements,too-many-locals
     def check_whitespace(  # noqa: max-complexity: 13
-        self, text: str
-    ) -> Tuple[str, str, str]:
+        self, text: str, key: str
+    ) -> tuple[str, str, str]:
         """Check spacing inside given jinja2 template string.
 
         We aim to match Python Black formatting rules.
 
         :returns: (string, string, string)  reformatted text, detailed error, error tag
         """
-        prev_token_type: str = ""
-        prev_value: str = ""
+
+        def cook(value: str, implicit: bool = False) -> str:
+            """Prepare an implicit string for jinja parsing when needed."""
+            if not implicit:
+                return value
+            if value.startswith("{{") and value.endswith("}}"):
+                # maybe we should make this an error?
+                return value
+            return f"{{{{ {value} }}}}"
+
+        def uncook(value: str, implicit: bool = False) -> str:
+            """Restore an string to original form when it was an implicit one."""
+            if not implicit:
+                return value
+            return value[3:-3]
+
         tokens = []
         details = ""
         begin_types = ("variable_begin", "comment_begin", "block_begin")
         end_types = ("variable_end", "comment_end", "block_end")
+        implicit = False
 
-        spaced_operators = {
-            "+",
-            "-",
-            "/",
-            "//",
-            "*",
-            "%",
-            "**",
-            "~",
-            "==",
-            "!=",
-            ">",
-            ">=",
-            "<",
-            "<=",
-            "=",
-            "|",
-        }
-        unspaced_operators = {"[", "]", "(", ")", "{", "}", "."}
-        space_folowed_operators = {",", ";", ":"}
-        pre_spaced_operators = spaced_operators
-        post_spaced_operators = spaced_operators | space_folowed_operators
-
-        def in_expression(tokens: List[Any]) -> str:
-            """Check if tokens represent an unfinished expression.
-
-            Returns last unclosed {[( or empty string.
-            """
-            opened: List[str] = []
-            pairs = {
-                ")": "(",
-                "]": "[",
-                "}": "{",
-            }
-            for _, token_type, value in tokens:  # reverse order
-                if token_type == "operator":
-                    if value in ("(", "[", "{"):
-                        opened.append(value)
-                    elif value in (")", "]", "}"):
-                        opened.remove(pairs[value])
-            if opened:
-                return opened[0]
-            return ""
+        # implicit templates do not have the {{ }} wrapping
+        if key in KEYWORDS_WITH_IMPLICIT_TEMPLATE:
+            implicit = True
+            text = cook(text, implicit=implicit)
 
+        expr_str = None
+        expr_type = None
+        verb_skipped = True
+        lineno = 1
         try:
-            previous_tokens = self.lex(text)
-        except jinja2.exceptions.TemplateSyntaxError as exc:
-            return "", str(exc.message), "invalid"
-
-        tokens = previous_tokens
-
-        # phase 1 : add missing whitespace
-        prev_token_type = ""
-        prev_value = ""
-        prev_lineno = 1
-        tokens = []
-        for lineno, token_type, value in previous_tokens:
-
-            if (
-                token_type in begin_types
-                and "-" in value
-                and prev_token_type == "data"
-                and not prev_value.endswith(" ")
-                and lineno == prev_lineno
-            ):
-                # "foo{%-" -> "foo {%-"
-                tokens[-1] = (tokens[-1][0], tokens[-1][1], tokens[-1][2] + " ")
-
-            if token_type in end_types and prev_token_type not in (
-                "whitespace",
-                "comment",
-            ):
-                tokens.append((lineno, "whitespace", " "))
-            elif prev_token_type in begin_types and token_type not in (
-                "whitespace",
-                "comment",
-            ):
-                tokens.append((lineno, "whitespace", " "))
-            if token_type in ("comment", "whitespace") and "\n" not in value:
-                # ensure comments/whitespace do not have more than one leading
-                # or trailing space in them, while not touching \n \r
-                stripped = " " + value.strip(" \t") + " "
-                if stripped == "  ":
-                    stripped = " "
-                if stripped != value:
-                    value = stripped
-            if (
-                token_type != "whitespace"
-                and token_type == "operator"
-                and value in pre_spaced_operators
-                and prev_token_type != "whitespace"
-            ):
-                tokens.append((lineno, "whitespace", " "))
-            if (
-                prev_token_type == "operator"
-                and prev_value in post_spaced_operators
-                and token_type != "whitespace"
-            ):
-                if not (
-                    prev_value == ":"
-                    and prev_token_type == "operator"
-                    and in_expression(tokens) == "["
-                ):
-                    avoid_spacing = False
-                    for token in tokens[:-1][::-1]:
-                        if token[1] in begin_types:
-                            avoid_spacing = True
-                            break
-                        if token[1] == "operator" and token[2] in (":", ""):
-                            avoid_spacing = True
-                            break
-                        if token[1] in ("operator", "integer", "string", "name"):
-                            avoid_spacing = False
-                            break
-                    if not avoid_spacing:
-                        tokens.append((lineno, "whitespace", " "))
-
-            tokens.append((lineno, token_type, value))
-            prev_token_type = token_type
-            prev_value = value
-            prev_lineno = lineno
-
-        # phase 2 : remove undesirable whitespace
-        prev_token_type = ""
-        prev_value = ""
-        previous_tokens = tokens
-        tokens = []
-        # pylint: disable=too-many-nested-blocks
-        for lineno, token_type, value in previous_tokens:
-
-            if prev_token_type == "whitespace" and "\n" not in prev_value:
+            for token in self.lex(text):
 
                 if (
-                    token_type == "operator"
-                    and value == ":"
-                    and tokens[-2][1] == "operator"
-                    and tokens[-2][2] == "["
+                    expr_type
+                    and expr_type.startswith("{%")
+                    and token.token_type in ("name", "whitespace")
+                    and not verb_skipped
                 ):
-                    tokens.pop()
-                elif token_type == "operator" and value in {
-                    *unspaced_operators,
-                    "=",
-                    ":",
-                }:
-                    # effective removal of whitespace
-                    if value in ("=", "]", ")", "}") and in_expression(tokens):
-                        tokens.pop()
-                    elif tokens[-2][1] == "operator" and tokens[-2][2] in {
-                        *unspaced_operators,
-                        "=",
-                    }:
-                        if tokens[-2][2] not in ("=", ",") or (
-                            tokens[-2][2] in ("=", ":") and in_expression(tokens)
-                        ):
-                            tokens.pop()
-                elif in_expression(tokens):
-                    if tokens[-2][1] == "operator":
-                        if tokens[-2][2] in (
-                            "=",
-                            "[",
-                            "{",
-                            "(",
-                        ):
-                            tokens.pop()
-                        elif tokens[-2][2] == ":" and in_expression(tokens) == "[":
-                            tokens.pop()
+                    # on {% blocks we do not take first word as part of the expression
+                    tokens.append(token)
+                    if token.token_type != "whitespace":
+                        verb_skipped = True
+                elif token.token_type in begin_types:
+                    tokens.append(token)
+                    expr_type = token.value  # such {#, {{, {%
+                    expr_str = ""
+                    verb_skipped = False
+                elif token.token_type in end_types and expr_str is not None:
+                    # process expression
+                    # pylint: disable=unsupported-membership-test
+                    if isinstance(expr_str, str) and "\n" in expr_str:
+                        raise NotImplementedError()
+                    expr_str = blacken(expr_str)
+                    if tokens[
+                        -1
+                    ].token_type != "whitespace" and not expr_str.startswith(" "):
+                        expr_str = " " + expr_str
+                    if not expr_str.endswith(" "):
+                        expr_str += " "
+                    tokens.append(Token(lineno, "data", expr_str))
+                    tokens.append(token)
+                    expr_str = None
+                    expr_type = None
+                elif expr_str is not None:
+                    expr_str += token.value
                 else:
-                    if tokens[-2][1] == "operator" and tokens[-2][2] in ("-", "+"):
-                        avoid_spacing = False
-                        for token in tokens[:-2][::-1]:
-                            if token[1] in begin_types:
-                                avoid_spacing = True
-                                break
-                            if token[1] in ("operator", "integer", "string", "name"):
-                                avoid_spacing = False
-                                break
-                        if avoid_spacing:
-                            tokens.pop()
+                    tokens.append(token)
+                lineno = token.lineno
 
-            tokens.append((lineno, token_type, value))
-            prev_token_type = token_type
-            prev_value = value
+        except jinja2.exceptions.TemplateSyntaxError as exc:
+            return "", str(exc.message), "invalid"
+        except (NotImplementedError, black.parsing.InvalidInput) as exc:
+            # black is not able to recognize all valid jinja2 templates, so we
+            # just ignore InvalidInput errors.
+            # NotImplementedError is raised internally for expressions with
+            # newlines, as we decided to not touch them yet.
+            # These both are documented as known limitations.
+            _logger.debug("Ignored jinja internal error %s", exc)
+            return uncook(text, implicit), "", "spacing"
 
         # finalize
         reformatted = self.unlex(tokens)
         failed = reformatted != text
+        reformatted = uncook(reformatted, implicit)
         details = (
             f"Jinja2 template rewrite recommendation: `{reformatted}`."
             if failed
@@ -329,6 +300,13 @@ class JinjaRule(AnsibleLintRule):
         return reformatted, details, "spacing"
 
 
+def blacken(text: str) -> str:
+    """Format Jinja2 template using black."""
+    return black.format_str(
+        text, mode=black.FileMode(line_length=sys.maxsize, string_normalization=False)
+    ).rstrip("\n")
+
+
 if "pytest" in sys.modules:  # noqa: C901
 
     import pytest
@@ -337,13 +315,13 @@ if "pytest" in sys.modules:  # noqa: C901
     from ansiblelint.runner import Runner  # pylint: disable=ungrouped-imports
 
     @pytest.fixture(name="error_expected_lines")
-    def fixture_error_expected_lines() -> List[int]:
+    def fixture_error_expected_lines() -> list[int]:
         """Return list of expected error lines."""
-        return [31, 34, 37, 40, 43, 46, 72, 83]
+        return [31, 34, 37, 40, 43, 46, 72]
 
     # 21 68
     @pytest.fixture(name="lint_error_lines")
-    def fixture_lint_error_lines() -> List[int]:
+    def fixture_lint_error_lines() -> list[int]:
         """Get VarHasSpacesRules linting results on test_playbook."""
         collection = RulesCollection()
         collection.register(JinjaRule())
@@ -352,7 +330,7 @@ if "pytest" in sys.modules:  # noqa: C901
         return list(map(lambda item: item.linenumber, results))
 
     def test_jinja_spacing_playbook(
-        error_expected_lines: List[int], lint_error_lines: List[int]
+        error_expected_lines: list[int], lint_error_lines: list[int]
     ) -> None:
         """Ensure that expected error lines are matching found linting error lines."""
         # list unexpected error lines or non-matching error lines
@@ -368,7 +346,7 @@ if "pytest" in sys.modules:  # noqa: C901
         lintable = Lintable("examples/playbooks/vars/jinja-spacing.yml")
         results = Runner(lintable, rules=collection).run()
 
-        error_expected_lineno = [14, 15, 16, 17, 18, 19, 32, 38]
+        error_expected_lineno = [14, 15, 16, 17, 18, 19, 32]
         assert len(results) == len(error_expected_lineno)
         for idx, err in enumerate(results):
             assert err.linenumber == error_expected_lineno[idx]
@@ -416,12 +394,12 @@ if "pytest" in sys.modules:  # noqa: C901
             ),
             pytest.param("{{foo(123)}}", "{{ foo(123) }}", "spacing", id="11"),
             pytest.param("{{ foo(a.b.c) }}", "{{ foo(a.b.c) }}", "spacing", id="12"),
-            pytest.param(
-                "{{ foo | bool else [ ] }}",
-                "{{ foo | bool else [] }}",
-                "spacing",
-                id="13",
-            ),
+            # pytest.param(
+            #     "{{ foo | bool else [ ] }}",
+            #     "{{ foo | bool else [] }}",
+            #     "spacing",
+            #     id="13",
+            # ),
             pytest.param(
                 "{{foo(x =['server_options'])}}",
                 "{{ foo(x=['server_options']) }}",
@@ -493,7 +471,7 @@ if "pytest" in sys.modules:  # noqa: C901
             pytest.param(
                 # "{% if a|int <= 8 -%} iptables {%- else -%} iptables-nft {%- endif %}",
                 "{% if a|int <= 8 -%} iptables {%- else -%} iptables-nft {%- endif %}",
-                "{% if a | int <= 8 -%} iptables {%- else -%} iptables-nft {%- endif %}",
+                "{% if a | int <= 8 -%} iptables{%- else -%} iptables-nft{%- endif %}",
                 "spacing",
                 id="29",
             ),
@@ -525,11 +503,92 @@ if "pytest" in sys.modules:  # noqa: C901
                 "spacing",
                 id="33",
             ),
+            pytest.param(
+                # negative array index
+                "{{ foo[-1] }}",
+                "{{ foo[-1] }}",
+                "spacing",
+                id="34",
+            ),
+            pytest.param(
+                # negative array index, repair
+                "{{ foo[- 1] }}",
+                "{{ foo[-1] }}",
+                "spacing",
+                id="35",
+            ),
+            pytest.param("{{ a +~'b' }}", "{{ a + ~'b' }}", "spacing", id="36"),
+            pytest.param(
+                "{{ (a[: -4] *~ b) }}", "{{ (a[:-4] * ~b) }}", "spacing", id="37"
+            ),
+            pytest.param("{{ [a,~ b] }}", "{{ [a, ~b] }}", "spacing", id="38"),
+            # Not supported yet due to being accepted by black:
+            pytest.param("{{ item.0.user }}", "{{ item.0.user }}", "spacing", id="39"),
+            # Not supported by back, while jinja allows ~ to be binary operator:
+            pytest.param("{{ a ~ b }}", "{{ a ~ b }}", "spacing", id="40"),
+            pytest.param(
+                "--format='{{'{{'}}.Size{{'}}'}}'",
+                "--format='{{ '{{' }}.Size{{ '}}' }}'",
+                "spacing",
+                id="41",
+            ),
+            pytest.param(
+                "{{ list_one + {{ list_two | max }} }}",
+                "{{ list_one + {{list_two | max}} }}",
+                "spacing",
+                id="42",
+            ),
         ),
     )
     def test_jinja(text: str, expected: str, tag: str) -> None:
         """Tests our ability to spot spacing errors inside jinja2 templates."""
         rule = JinjaRule()
-        reformatted, details, returned_tag = rule.check_whitespace(text)
+
+        reformatted, details, returned_tag = rule.check_whitespace(text, key="name")
         assert tag == returned_tag, details
         assert expected == reformatted
+
+    @pytest.mark.parametrize(
+        ("text", "expected", "tag"),
+        (
+            pytest.param(
+                "1+2",
+                "1 + 2",
+                "spacing",
+                id="0",
+            ),
+            pytest.param(
+                "- 1",
+                "-1",
+                "spacing",
+                id="1",
+            ),
+            # Ensure that we do not choke with double templating on implicit
+            # and instead we remove them braces.
+            pytest.param("{{ o | bool }}", "o | bool", "spacing", id="2"),
+        ),
+    )
+    def test_jinja_implicit(text: str, expected: str, tag: str) -> None:
+        """Tests our ability to spot spacing errors implicit jinja2 templates."""
+        rule = JinjaRule()
+        reformatted, details, returned_tag = rule.check_whitespace(text, key="when")
+        assert tag == returned_tag, details
+        assert expected == reformatted
+
+    def test_jinja_invalid() -> None:
+        """Tests our ability to spot spacing errors inside jinja2 templates."""
+        collection = RulesCollection()
+        collection.register(JinjaRule())
+        success = "examples/playbooks/rule-jinja-invalid.yml"
+        errs = Runner(success, rules=collection).run()
+        assert len(errs) == 1
+        assert errs[0].tag == "jinja[invalid]"
+        assert errs[0].rule.id == "jinja"
+
+    def test_jinja_valid() -> None:
+        """Tests our ability to parse jinja, even when variables may not be defined."""
+        collection = RulesCollection()
+        collection.register(JinjaRule())
+        success = "examples/playbooks/rule-jinja-valid.yml"
+        errs = Runner(success, rules=collection).run()
+        assert len(errs) == 0
